# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""Oauth authentication plugins.

.. warning::
    This module requires installation of an extra package (`oauthlib`)
    not installed by default. Without the extra package an import error will
    occur. The extra package can be installed using::

      $ pip install keystoneauth['oauth1']
"""

import logging
import typing as ty

try:
    # explicitly re-export symbol
    # https://mypy.readthedocs.io/en/stable/command_line.html#cmdoption-mypy-no-implicit-reexport
    from oauthlib import oauth1 as oauth1
except ImportError:
    oauth1 = None

from keystoneauth1.identity import v3
from keystoneauth1 import session as ks_session

__all__ = ('OAuth1Method', 'OAuth1')

LOG = logging.getLogger(__name__)


class OAuth1Method(v3.AuthMethod):
    """OAuth based authentication method.

    :param string access_key: Access token key.
    :param string access_secret: Access token secret.
    :param string consumer_key: Consumer key.
    :param string consumer_secret: Consumer secret.
    """

    access_key: str
    access_secret: str
    consumer_key: str
    consumer_secret: str

    def __init__(
        self,
        *,
        access_key: str,
        access_secret: str,
        consumer_key: str,
        consumer_secret: str,
    ) -> None:
        self.access_key = access_key
        self.access_secret = access_secret
        self.consumer_key = consumer_key
        self.consumer_secret = consumer_secret

    def get_auth_data(
        self,
        session: ks_session.Session,
        auth: v3.Auth,
        headers: dict[str, str],
        request_kwargs: dict[str, object],
    ) -> tuple[None, None] | tuple[str, ty.Mapping[str, object]]:
        # Add the oauth specific content into the headers
        oauth_client = oauth1.Client(
            self.consumer_key,
            client_secret=self.consumer_secret,
            resource_owner_key=self.access_key,
            resource_owner_secret=self.access_secret,
            signature_method=oauth1.SIGNATURE_HMAC,
        )

        o_url, o_headers, o_body = oauth_client.sign(
            auth.token_url, http_method='POST'
        )
        headers.update(o_headers)

        return 'oauth1', {}

    def get_cache_id_elements(self) -> dict[str, str | None]:
        return {
            'oauth1_access_key': self.access_key,
            'oauth1_access_secret': self.access_secret,
            'oauth1_consumer_key': self.consumer_key,
            'oauth1_consumer_secret': self.consumer_secret,
        }


class OAuth1(v3.Auth):
    _auth_method_class = OAuth1Method

    def __init__(
        self,
        auth_url: str,
        consumer_key: str,
        consumer_secret: str,
        access_key: str,
        access_secret: str,
        *,
        unscoped: bool = False,
        trust_id: str | None = None,
        system_scope: str | None = None,
        domain_id: str | None = None,
        domain_name: str | None = None,
        project_id: str | None = None,
        project_name: str | None = None,
        project_domain_id: str | None = None,
        project_domain_name: str | None = None,
        reauthenticate: bool = True,
        include_catalog: bool = True,
    ) -> None:
        method = self._auth_method_class(
            consumer_key=consumer_key,
            consumer_secret=consumer_secret,
            access_key=access_key,
            access_secret=access_secret,
        )
        super().__init__(
            auth_url,
            [method],
            unscoped=unscoped,
            trust_id=trust_id,
            system_scope=system_scope,
            domain_id=domain_id,
            domain_name=domain_name,
            project_id=project_id,
            project_name=project_name,
            project_domain_id=project_domain_id,
            project_domain_name=project_domain_name,
            reauthenticate=reauthenticate,
            include_catalog=include_catalog,
        )

        if self.has_scope_parameters:
            LOG.warning(
                'Scoping parameters such as a project were provided '
                'to the OAuth1 plugin. Because OAuth1 access is '
                'always scoped to a project these will be ignored by '
                'the identity server'
            )



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import uuid

from oauthlib import oauth1
from testtools import matchers

from keystoneauth1.extras import oauth1 as ksa_oauth1
from keystoneauth1 import fixture
from keystoneauth1 import session
from keystoneauth1.tests.unit import utils as test_utils


class OAuth1AuthTests(test_utils.TestCase):
    TEST_ROOT_URL = 'http://127.0.0.1:5000/'
    TEST_URL = '{}{}'.format(TEST_ROOT_URL, 'v3')
    TEST_TOKEN = uuid.uuid4().hex

    def stub_auth(self, subject_token=None, **kwargs):
        if not subject_token:
            subject_token = self.TEST_TOKEN

        self.stub_url(
            'POST',
            ['auth', 'tokens'],
            headers={'X-Subject-Token': subject_token},
            **kwargs,
        )

    def _validate_oauth_headers(self, auth_header, oauth_client):
        """Validate data in the headers.

        Assert that the data in the headers matches the data
        that is produced from oauthlib.
        """
        self.assertThat(auth_header, matchers.StartsWith('OAuth '))
        parameters = dict(
            oauth1.rfc5849.utils.parse_authorization_header(auth_header)
        )

        self.assertEqual('HMAC-SHA1', parameters['oauth_signature_method'])
        self.assertEqual('1.0', parameters['oauth_version'])
        self.assertIsInstance(parameters['oauth_nonce'], str)
        self.assertEqual(
            oauth_client.client_key, parameters['oauth_consumer_key']
        )
        if oauth_client.resource_owner_key:
            self.assertEqual(
                oauth_client.resource_owner_key, parameters['oauth_token']
            )
        if oauth_client.verifier:
            self.assertEqual(
                oauth_client.verifier, parameters['oauth_verifier']
            )
        if oauth_client.callback_uri:
            self.assertEqual(
                oauth_client.callback_uri, parameters['oauth_callback']
            )
        return parameters

    def test_oauth_authenticate_success(self):
        consumer_key = uuid.uuid4().hex
        consumer_secret = uuid.uuid4().hex
        access_key = uuid.uuid4().hex
        access_secret = uuid.uuid4().hex

        oauth_token = fixture.V3Token(
            methods=['oauth1'],
            oauth_consumer_id=consumer_key,
            oauth_access_token_id=access_key,
        )
        oauth_token.set_project_scope()

        self.stub_auth(json=oauth_token)

        a = ksa_oauth1.V3OAuth1(
            self.TEST_URL,
            consumer_key=consumer_key,
            consumer_secret=consumer_secret,
            access_key=access_key,
            access_secret=access_secret,
        )

        s = session.Session(auth=a)
        t = s.get_token()

        self.assertEqual(self.TEST_TOKEN, t)

        OAUTH_REQUEST_BODY = {
            "auth": {"identity": {"methods": ["oauth1"], "oauth1": {}}}
        }

        self.assertRequestBodyIs(json=OAUTH_REQUEST_BODY)

        # Assert that the headers have the same oauthlib data
        req_headers = self.requests_mock.last_request.headers
        oauth_client = oauth1.Client(
            consumer_key,
            client_secret=consumer_secret,
            resource_owner_key=access_key,
            resource_owner_secret=access_secret,
            signature_method=oauth1.SIGNATURE_HMAC,
        )
        self._validate_oauth_headers(
            req_headers['Authorization'], oauth_client
        )

    def test_warning_dual_scope(self):
        ksa_oauth1.V3OAuth1(
            self.TEST_URL,
            consumer_key=uuid.uuid4().hex,
            consumer_secret=uuid.uuid4().hex,
            access_key=uuid.uuid4().hex,
            access_secret=uuid.uuid4().hex,
            project_id=uuid.uuid4().hex,
        )

        self.assertIn('ignored by the identity server', self.logger.output)



# Copyright 2012 Nebula, Inc.
#
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import datetime
import typing as ty

import requests

from keystoneauth1 import _utils as utils
from keystoneauth1.access import service_catalog
from keystoneauth1.access import service_providers
from keystoneauth1.access import types

# gap, in seconds, to determine whether the given token is about to expire
STALE_TOKEN_DURATION = 30

__all__ = ('AccessInfo', 'AccessInfoV2', 'AccessInfoV3', 'create')


def create(
    resp: requests.Response | None = None,
    body: dict[str, object] | None = None,
    auth_token: str | None = None,
) -> 'AccessInfo':
    if resp and not body:
        body = resp.json()

    assert body is not None  # nosec B101

    if 'token' in body:
        if resp and not auth_token:
            auth_token = resp.headers.get('X-Subject-Token')

        return AccessInfoV3(body, auth_token)
    elif 'access' in body:
        return AccessInfoV2(body, auth_token)

    raise ValueError('Unrecognized auth response')


class AccessInfo:
    """Encapsulates a raw authentication token from keystone.

    Provides helper methods for extracting useful values from that token.
    """

    _service_catalog_class: type[service_catalog.ServiceCatalog]
    _data: ty.Any

    def __init__(self, body: dict[str, ty.Any], auth_token: str | None = None):
        self._data = body
        self._auth_token = auth_token
        self._service_catalog: service_catalog.ServiceCatalog | None = None
        self._service_providers: service_providers.ServiceProviders | None = (
            None
        )

    @property
    def service_catalog(self) -> service_catalog.ServiceCatalog:
        if not self._service_catalog:
            self._service_catalog = self._service_catalog_class.from_token(
                self._data
            )

        return self._service_catalog

    def will_expire_soon(
        self, stale_duration: int = STALE_TOKEN_DURATION
    ) -> bool:
        """Determine if expiration is about to occur.

        :returns: true if expiration is within the given duration
        :rtype: boolean
        """
        if self.expires is None:
            return False

        norm_expires = utils.normalize_time(self.expires)
        # (gyee) should we move auth_token.will_expire_soon() to timeutils
        # instead of duplicating code here?
        soon = utils.from_utcnow(seconds=stale_duration)
        return norm_expires < soon

    def has_service_catalog(self) -> bool:
        """Return true if the auth token has a service catalog.

        :returns: boolean
        """
        raise NotImplementedError()

    @property
    def auth_token(self) -> str | None:
        """Return the token_id associated with the auth request.

        To be used in headers for authenticating OpenStack API requests.

        :returns: str
        """
        return self._auth_token

    @property
    def expires(self) -> datetime.datetime | None:
        """Return the token expiration (as datetime object).

        :returns: datetime
        """
        raise NotImplementedError()

    @property
    def issued(self) -> datetime.datetime | None:
        """Return the token issue time (as datetime object).

        :returns: datetime
        """
        raise NotImplementedError()

    @property
    def username(self) -> str | None:
        """Return the username associated with the auth request.

        Follows the pattern defined in the V2 API of first looking for 'name',
        returning that if available, and falling back to 'username' if name
        is unavailable.

        :returns: str
        """
        raise NotImplementedError()

    @property
    def user_id(self) -> str | None:
        """Return the user id associated with the auth request.

        :returns: str
        """
        raise NotImplementedError()

    @property
    def user_domain_id(self) -> str | None:
        """Return the user's domain id associated with the auth request.

        :returns: str
        """
        raise NotImplementedError()

    @property
    def user_domain_name(self) -> str | None:
        """Return the user's domain name associated with the auth request.

        :returns: str
        """
        raise NotImplementedError()

    @property
    def role_ids(self) -> list[str] | None:
        """Return a list of user's role ids associated with the auth request.

        :returns: a list of strings of role ids
        """
        raise NotImplementedError()

    @property
    def role_names(self) -> list[str] | None:
        """Return a list of user's role names associated with the auth request.

        :returns: a list of strings of role names
        """
        raise NotImplementedError()

    @property
    def domain_name(self) -> str | None:
        """Return the domain name associated with the auth request.

        :returns: str or None (if no domain associated with the token)
        """
        raise NotImplementedError()

    @property
    def domain_id(self) -> str | None:
        """Return the domain id associated with the auth request.

        :returns: str or None (if no domain associated with the token)
        """
        raise NotImplementedError()

    @property
    def project_name(self) -> str | None:
        """Return the project name associated with the auth request.

        :returns: str or None (if no project associated with the token)
        """
        raise NotImplementedError()

    @property
    def tenant_name(self) -> str | None:
        """Synonym for project_name."""
        return self.project_name

    @property
    def scoped(self) -> bool:
        """Return true if the auth token was scoped.

        Returns true if scoped to a tenant(project) or domain,
        and contains a populated service catalog.

        This is deprecated, use project_scoped instead.

        :returns: bool
        """
        return self.project_scoped or self.domain_scoped or self.system_scoped

    @property
    def project_scoped(self) -> bool:
        """Return true if the auth token was scoped to a tenant (project).

        :returns: bool
        """
        return bool(self.project_id)

    @property
    def domain_scoped(self) -> bool:
        """Return true if the auth token was scoped to a domain.

        :returns: bool
        """
        raise NotImplementedError()

    @property
    def system_scoped(self) -> bool:
        """Return true if the auth token was scoped to the system.

        :returns: bool
        """
        raise NotImplementedError()

    @property
    def trust_id(self) -> str | None:
        """Return the trust id associated with the auth request.

        :returns: str or None (if no trust associated with the token)
        """
        raise NotImplementedError()

    @property
    def trust_scoped(self) -> bool:
        """Return true if the auth token was scoped from a delegated trust.

        The trust delegation is via the OS-TRUST v3 extension.

        :returns: bool
        """
        raise NotImplementedError()

    @property
    def trustee_user_id(self) -> str | None:
        """Return the trustee user id associated with a trust.

        :returns: str or None (if no trust associated with the token)
        """
        raise NotImplementedError()

    @property
    def trustor_user_id(self) -> str | None:
        """Return the trustor user id associated with a trust.

        :returns: str or None (if no trust associated with the token)
        """
        raise NotImplementedError()

    @property
    def project_id(self) -> str | None:
        """Return the project ID associated with the auth request.

        This returns None if the auth token wasn't scoped to a project.

        :returns: str or None (if no project associated with the token)
        """
        raise NotImplementedError()

    @property
    def tenant_id(self) -> str | None:
        """Synonym for project_id."""
        return self.project_id

    @property
    def project_domain_id(self) -> str | None:
        """Return the project's domain id associated with the auth request.

        :returns: str
        """
        raise NotImplementedError()

    @property
    def project_domain_name(self) -> str | None:
        """Return the project's domain name associated with the auth request.

        :returns: str
        """
        raise NotImplementedError()

    @property
    def oauth_access_token_id(self) -> str | None:
        """Return the access token ID if OAuth authentication used.

        :returns: str or None.
        """
        raise NotImplementedError()

    @property
    def oauth_consumer_id(self) -> str | None:
        """Return the consumer ID if OAuth authentication used.

        :returns: str or None.
        """
        raise NotImplementedError()

    @property
    def is_federated(self) -> bool:
        """Return true if federation was used to get the token.

        :returns: boolean
        """
        raise NotImplementedError()

    @property
    def is_admin_project(self) -> bool:
        """Return true if the current project scope is the admin project.

        For backwards compatibility purposes if there is nothing specified in
        the token we always assume we are in the admin project, so this will
        default to True.

        :returns boolean
        """
        raise NotImplementedError()

    @property
    def audit_id(self) -> str | None:
        """Return the audit ID if present.

        :returns: str or None.
        """
        raise NotImplementedError()

    @property
    def audit_chain_id(self) -> str | None:
        """Return the audit chain ID if present.

        In the event that a token was rescoped then this ID will be the
        :py:attr:`audit_id` of the initial token. Returns None if no value
        present.

        :returns: str or None.
        """
        raise NotImplementedError()

    @property
    def initial_audit_id(self) -> str | None:
        """The audit ID of the initially requested token.

        This is the :py:attr:`audit_chain_id` if present or the
        :py:attr:`audit_id`.
        """
        return self.audit_chain_id or self.audit_id

    @property
    def service_providers(self) -> service_providers.ServiceProviders | None:
        """Return an object representing the list of trusted service providers.

        Used for Keystone2Keystone federating-out.

        :returns: :py:class:`keystoneauth1.service_providers.ServiceProviders`
                  or None
        """
        raise NotImplementedError()

    @property
    def bind(self) -> dict[str, ty.Any] | None:
        """Information about external mechanisms the token is bound to.

        If a token is bound to an external authentication mechanism it can only
        be used in conjunction with that mechanism. For example if bound to a
        kerberos principal it may only be accepted if there is also kerberos
        authentication performed on the request.

        :returns: A dictionary or None. The key will be the bind type the value
                  is a dictionary that is specific to the format of the bind
                  type. Returns None if there is no bind information in the
                  token.
        """
        raise NotImplementedError()

    @property
    def project_is_domain(self) -> bool | None:
        """Return if a project act as a domain.

        :returns: bool
        """
        raise NotImplementedError()


class AccessInfoV2(AccessInfo):
    """An object for encapsulating raw v2 auth token from identity service."""

    version = 'v2.0'
    _service_catalog_class = service_catalog.ServiceCatalogV2
    _data: types.TokenResponseV2

    def has_service_catalog(self) -> bool:
        return 'serviceCatalog' in self._data['access']

    @property
    def _token(self) -> types.TokenV2:
        return self._data['access']['token']

    @property
    def auth_token(self) -> str | None:
        set_token = super().auth_token
        return set_token or self._token.get('id')

    @property
    def expires(self) -> datetime.datetime | None:
        return utils.parse_isotime(self._token['expires'])

    @property
    def issued(self) -> datetime.datetime | None:
        return utils.parse_isotime(self._token['issued_at'])

    @property
    def _user(self) -> types.UserV2:
        return self._data['access']['user']

    @property
    def username(self) -> str | None:
        return self._user.get('name') or self._user.get('username')

    @property
    def user_id(self) -> str | None:
        return self._user.get('id')

    @property
    def user_domain_id(self) -> str | None:
        return None

    @property
    def user_domain_name(self) -> str | None:
        return None

    @property
    def role_ids(self) -> list[str] | None:
        metadata = self._data['access'].get('metadata', {})
        return metadata.get('roles', [])

    @property
    def role_names(self) -> list[str] | None:
        return [r['name'] for r in self._user.get('roles', [])]

    @property
    def domain_name(self) -> str | None:
        return None

    @property
    def domain_id(self) -> str | None:
        return None

    @property
    def project_name(self) -> str | None:
        if 'tenant' in self._token:
            return self._token['tenant'].get('name')

        # pre grizzly
        if 'tenantName' in self._user:
            return self._user['tenantName']

        # pre diablo, keystone only provided a tenantId
        if 'tenantId' in self._token:
            return self._token['tenantId']  # type: ignore

        return None

    @property
    def domain_scoped(self) -> bool:
        return False

    @property
    def system_scoped(self) -> bool:
        return False

    @property
    def _trust(self) -> types.TrustV2 | None:
        return self._data['access'].get('trust')

    @property
    def trust_id(self) -> str | None:
        return self._trust and self._trust['id']

    @property
    def trust_scoped(self) -> bool:
        return bool(self._trust)

    @property
    def trustee_user_id(self) -> str | None:
        return self._trust and self._trust['trustee_user_id']

    @property
    def trustor_user_id(self) -> str | None:
        # this information is not available in the v2 token bug: #1331882
        return None

    @property
    def project_id(self) -> str | None:
        if 'tenant' in self._token:
            return self._token['tenant'].get('id')

        # pre grizzly
        if 'tenantId' in self._user:
            return self._user['tenantId']

        # pre diablo, keystone only provided a tenantId
        if 'tenantId' in self._token:
            return self._token['tenantId']  # type: ignore

        return None

    @property
    def project_is_domain(self) -> bool | None:
        return False

    @property
    def project_domain_id(self) -> str | None:
        return None

    @property
    def project_domain_name(self) -> str | None:
        return None

    @property
    def oauth_access_token_id(self) -> str | None:
        return None

    @property
    def oauth_consumer_id(self) -> str | None:
        return None

    @property
    def is_federated(self) -> bool:
        return False

    @property
    def is_admin_project(self) -> bool:
        return True

    @property
    def audit_id(self) -> str | None:
        try:
            return self._token.get('audit_ids', [])[0]
        except IndexError:
            return None

    @property
    def audit_chain_id(self) -> str | None:
        try:
            return self._token.get('audit_ids', [])[1]
        except IndexError:
            return None

    @property
    def service_providers(self) -> service_providers.ServiceProviders | None:
        return None

    @property
    def bind(self) -> dict[str, ty.Any] | None:
        return self._token.get('bind')


class AccessInfoV3(AccessInfo):
    """An object encapsulating raw v3 auth token from identity service."""

    version = 'v3'
    _service_catalog_class = service_catalog.ServiceCatalogV3
    _data: types.TokenResponseV3

    @property
    def _token(self) -> types.TokenV3:
        return self._data['token']

    def has_service_catalog(self) -> bool:
        return 'catalog' in self._token

    @property
    def expires(self) -> datetime.datetime | None:
        return utils.parse_isotime(self._token['expires_at'])

    @property
    def issued(self) -> datetime.datetime | None:
        return utils.parse_isotime(self._token['issued_at'])

    @property
    def _user(self) -> types.UserV3:
        return self._token['user']

    @property
    def username(self) -> str | None:
        return self._user['name']

    @property
    def user_id(self) -> str | None:
        return self._user['id']

    @property
    def _user_domain(self) -> types.UserDomainV3:
        return self._user['domain']

    @property
    def user_domain_id(self) -> str | None:
        return self._user['domain']['id']

    @property
    def user_domain_name(self) -> str | None:
        return self._user['domain']['name']

    @property
    def role_ids(self) -> list[str] | None:
        return [r['id'] for r in self._token.get('roles', [])]

    @property
    def role_names(self) -> list[str] | None:
        return [r['name'] for r in self._token.get('roles', [])]

    @property
    def system(self) -> types.SystemV3 | None:
        return self._token.get('system')

    @property
    def _domain(self) -> types.DomainV3 | None:
        # only present for domain-scoped tokens
        return self._token.get('domain')

    @property
    def domain_name(self) -> str | None:
        return self._domain and self._domain['name']

    @property
    def domain_id(self) -> str | None:
        return self._domain and self._domain['id']

    @property
    def _project(self) -> types.ProjectV3 | None:
        # only present for project-scoped tokens
        return self._token.get('project')

    @property
    def project_id(self) -> str | None:
        return self._project and self._project.get('id')

    @property
    def project_name(self) -> str | None:
        return self._project and self._project.get('name')

    @property
    def project_is_domain(self) -> bool | None:
        return self._token.get('is_domain')

    @property
    def _project_domain(self) -> types.ProjectDomainV3 | None:
        return self._project and self._project.get('domain')

    @property
    def project_domain_id(self) -> str | None:
        return self._project_domain and self._project_domain['id']

    @property
    def project_domain_name(self) -> str | None:
        return self._project_domain and self._project_domain['name']

    @property
    def domain_scoped(self) -> bool:
        return bool(self._domain)

    @property
    def system_scoped(self) -> bool:
        return bool(self._token.get('system', {}).get('all'))

    @property
    def _trust(self) -> types.TrustV3 | None:
        # only present for trust-scoped tokens
        return self._token.get('OS-TRUST:trust')

    @property
    def trust_id(self) -> str | None:
        return self._trust and self._trust['id']

    @property
    def trust_scoped(self) -> bool:
        return bool(self._trust)

    @property
    def trustee_user_id(self) -> str | None:
        return self._trust and self._trust['trustee_user']['id']

    @property
    def trustor_user_id(self) -> str | None:
        return self._trust and self._trust['trustor_user']['id']

    # TODO(stephenfin): Should this be private like every other high-level
    # accessor? As this stands, it can raise KeyError
    @property
    def application_credential(self) -> types.ApplicationCredentialV3:
        return self._token['application_credential']

    @property
    def _application_credential(self) -> types.ApplicationCredentialV3 | None:
        # only present if user has authenticated with application credentials
        return self._token.get('application_credential')

    @property
    def application_credential_id(self) -> str | None:
        return (
            self._application_credential
            and self._application_credential.get('id')
        )

    @property
    def application_credential_access_rules(
        self,
    ) -> list[types.ApplicationCredentialAccessRuleV3] | None:
        return (
            self._application_credential
            and self._application_credential.get('access_rules')
        )

    @property
    def _oauth(self) -> types.OAuth1V3 | None:
        # only present if user has authenticated with OAuth1
        return self._token.get('OS-OAUTH1')

    @property
    def oauth_access_token_id(self) -> str | None:
        return self._oauth and self._oauth.get('access_token_id')

    @property
    def oauth_consumer_id(self) -> str | None:
        return self._oauth and self._oauth.get('consumer_id')

    @property
    def is_federated(self) -> bool:
        return 'OS-FEDERATION' in self._user

    @property
    def is_admin_project(self) -> bool:
        return bool(self._token.get('is_admin_project', True))

    @property
    def audit_id(self) -> str | None:
        ret = self._token.get('audit_ids', [])
        return ret[0] if ret else None

    @property
    def audit_chain_id(self) -> str | None:
        ret = self._token.get('audit_ids', [])
        return ret[1] if len(ret) > 1 else None

    @property
    def service_providers(self) -> service_providers.ServiceProviders | None:
        if not self._service_providers:
            self._service_providers = (
                service_providers.ServiceProviders.from_token(self._data)
            )

        return self._service_providers

    @property
    def bind(self) -> dict[str, ty.Any] | None:
        return self._token.get('bind')

    # TODO(stephenfin): Should this be private like every other high-level
    # accessor? As this stands, it can raise KeyError
    @property
    def oauth2_credential(self) -> types.OAuth2V3:
        return self._token['oauth2_credential']

    @property
    def _oauth2_credential(self) -> types.OAuth2V3 | None:
        return self._token.get('oauth2_credential')

    @property
    def oauth2_credential_thumbprint(self) -> str | None:
        return self._oauth2_credential and self._oauth2_credential.get(
            'x5t#S256'
        )



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import abc
import typing as ty

from keystoneauth1 import _utils as utils
from keystoneauth1 import plugin

if ty.TYPE_CHECKING:
    from keystoneauth1 import session as ks_session

LOG = utils.get_logger(__name__)


class TokenlessAuth(plugin.BaseAuthPlugin, metaclass=abc.ABCMeta):
    """A plugin for authenticating with Tokenless Auth.

    This is for Tokenless Authentication. Scoped information
    like domain name and project ID will be passed in the headers and
    token validation request will be authenticated based on
    the provided HTTPS certificate along with the scope information.
    """

    def __init__(
        self,
        auth_url: str,
        domain_id: str | None = None,
        domain_name: str | None = None,
        project_id: str | None = None,
        project_name: str | None = None,
        project_domain_id: str | None = None,
        project_domain_name: str | None = None,
    ):
        """A init method for TokenlessAuth.

        :param string auth_url: Identity service endpoint for authentication.
                                The URL must include a version or any request
                                will result in a 404 NotFound error.
        :param string domain_id: Domain ID for domain scoping.
        :param string domain_name: Domain name for domain scoping.
        :param string project_id: Project ID for project scoping.
        :param string project_name: Project name for project scoping.
        :param string project_domain_id: Project's domain ID for project.
        :param string project_domain_name: Project's domain name for project.
        """
        self.auth_url = auth_url
        self.domain_id = domain_id
        self.domain_name = domain_name
        self.project_id = project_id
        self.project_name = project_name
        self.project_domain_id = project_domain_id
        self.project_domain_name = project_domain_name

    def get_headers(
        self, session: 'ks_session.Session'
    ) -> dict[str, str] | None:
        """Fetch authentication headers for message.

        This is to override the default get_headers method to provide
        tokenless auth scope headers if token is not provided in the
        session.

        :param session: The session object that the auth_plugin belongs to.
        :type session: keystoneauth1.session.Session

        :returns: Headers that are set to authenticate a message or None for
                  failure. Note that when checking this value that the empty
                  dict is a valid, non-failure response.
        :rtype: dict
        """
        scope_headers = {}
        if self.project_id:
            scope_headers['X-Project-Id'] = self.project_id
        elif self.project_name:
            scope_headers['X-Project-Name'] = self.project_name
            if self.project_domain_id:
                scope_headers['X-Project-Domain-Id'] = self.project_domain_id
            elif self.project_domain_name:
                scope_headers['X-Project-Domain-Name'] = (
                    self.project_domain_name
                )
            else:
                LOG.warning(
                    'Neither Project Domain ID nor Project Domain Name was '
                    'provided.'
                )
                return None
        elif self.domain_id:
            scope_headers['X-Domain-Id'] = self.domain_id
        elif self.domain_name:
            scope_headers['X-Domain-Name'] = self.domain_name
        else:
            LOG.warning('Neither Project nor Domain scope was provided.')
            return None
        return scope_headers

    def get_endpoint(
        self,
        session: 'ks_session.Session',
        service_type: str | None = None,
        **kwargs: ty.Any,
    ) -> str | None:
        """Return a valid endpoint for a service.

        :param session: A session object that can be used for communication.
        :type session: keystoneauth1.session.Session
        :param string service_type: The type of service to lookup the endpoint
                                    for. This plugin will return None (failure)
                                    if service_type is not provided.
        :return: A valid endpoint URL or None if not available.
        :rtype: string or None
        """
        if service_type is plugin.AUTH_INTERFACE or (
            service_type and service_type.lower() == 'identity'
        ):
            return self.auth_url

        return None



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import uuid

from keystoneauth1 import loading
from keystoneauth1.tests.unit import utils as test_utils


class OAuth1LoadingTests(test_utils.TestCase):
    def setUp(self):
        super().setUp()
        self.auth_url = uuid.uuid4().hex

    def create(self, **kwargs):
        kwargs.setdefault('auth_url', self.auth_url)
        loader = loading.get_plugin_loader('v3oauth1')
        return loader.load_from_options(**kwargs)

    def test_basic(self):
        access_key = uuid.uuid4().hex
        access_secret = uuid.uuid4().hex
        consumer_key = uuid.uuid4().hex
        consumer_secret = uuid.uuid4().hex

        p = self.create(
            access_key=access_key,
            access_secret=access_secret,
            consumer_key=consumer_key,
            consumer_secret=consumer_secret,
        )

        oauth_method = p.auth_methods[0]

        self.assertEqual(self.auth_url, p.auth_url)
        self.assertEqual(access_key, oauth_method.access_key)
        self.assertEqual(access_secret, oauth_method.access_secret)
        self.assertEqual(consumer_key, oauth_method.consumer_key)
        self.assertEqual(consumer_secret, oauth_method.consumer_secret)

    def test_options(self):
        options = loading.get_plugin_loader('v3oauth1').get_options()

        self.assertEqual(
            {o.name for o in options},
            {
                'auth-url',
                'access-key',
                'access-secret',
                'consumer-key',
                'consumer-secret',
            },
        )



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import typing as ty

from keystoneauth1 import _utils as utils

__all__ = ('DiscoveryList', 'V2Discovery', 'V3Discovery', 'VersionDiscovery')

_DEFAULT_DAYS_AGO = 30


class DiscoveryBase(dict[str, ty.Any]):
    """The basic version discovery structure.

    All version discovery elements should have access to these values.

    :param string id: The version id for this version entry.
    :param string status: The status of this entry.
    :param DateTime updated: When the API was last updated.
    """

    def __init__(self, id, status=None, updated=None):
        super().__init__()

        self.id = id
        self.status = status or 'stable'
        self.updated = updated or utils.before_utcnow(days=_DEFAULT_DAYS_AGO)

    @property
    def id(self):
        return self.get('id')

    @id.setter
    def id(self, value):
        self['id'] = value

    @property
    def status(self):
        return self.get('status')

    @status.setter
    def status(self, value):
        self['status'] = value

    @property
    def links(self):
        return self.setdefault('links', [])

    @property
    def updated_str(self):
        return self.get('updated')

    @updated_str.setter
    def updated_str(self, value):
        self['updated'] = value

    @property
    def updated(self):
        return utils.parse_isotime(self.updated_str)

    @updated.setter
    def updated(self, value):
        self.updated_str = value.isoformat()

    def add_link(self, href, rel='self', type=None):
        link = {'href': href, 'rel': rel}
        if type:
            link['type'] = type
        self.links.append(link)
        return link

    @property
    def media_types(self):
        return self.setdefault('media-types', [])

    def add_media_type(self, base, type):
        mt = {'base': base, 'type': type}
        self.media_types.append(mt)
        return mt


class VersionDiscovery(DiscoveryBase):
    """A Version element for non-keystone services without microversions.

    Provides some default values and helper methods for creating a microversion
    endpoint version structure. Clients should use this instead of creating
    their own structures.

    :param string href: The url that this entry should point to.
    :param string id: The version id that should be reported.
    """

    def __init__(self, href, id, **kwargs):
        super().__init__(id, **kwargs)

        self.add_link(href)


class MicroversionDiscovery(DiscoveryBase):
    """A Version element that has microversions.

    Provides some default values and helper methods for creating a microversion
    endpoint version structure. Clients should use this instead of creating
    their own structures.

    :param string href: The url that this entry should point to.
    :param string id: The version id that should be reported.
    :param string min_version: The minimum supported microversion. (optional)
    :param string max_version: The maximum supported microversion. (optional)
    """

    def __init__(self, href, id, min_version='', max_version='', **kwargs):
        super().__init__(id, **kwargs)

        self.add_link(href)

        self.min_version = min_version
        self.max_version = max_version

    @property
    def min_version(self):
        return self.get('min_version')

    @min_version.setter
    def min_version(self, value):
        self['min_version'] = value

    @property
    def max_version(self):
        return self.get('max_version')

    @max_version.setter
    def max_version(self, value):
        self['max_version'] = value


class NovaMicroversionDiscovery(DiscoveryBase):
    """A Version element with nova-style microversions.

    Provides some default values and helper methods for creating a microversion
    endpoint version structure. Clients should use this instead of creating
    their own structures.

    :param href: The url that this entry should point to.
    :param string id: The version id that should be reported.
    :param string min_version: The minimum microversion supported. (optional)
    :param string version: The maximum microversion supported. (optional)
    """

    def __init__(self, href, id, min_version=None, version=None, **kwargs):
        super().__init__(id, **kwargs)

        self.add_link(href)

        self.min_version = min_version
        self.version = version

    @property
    def min_version(self):
        return self.get('min_version')

    @min_version.setter
    def min_version(self, value):
        if value:
            self['min_version'] = value

    @property
    def version(self):
        return self.get('version')

    @version.setter
    def version(self, value):
        if value:
            self['version'] = value


class V2Discovery(DiscoveryBase):
    """A Version element for a V2 identity service endpoint.

    Provides some default values and helper methods for creating a v2.0
    endpoint version structure. Clients should use this instead of creating
    their own structures.

    :param string href: The url that this entry should point to.
    :param string id: The version id that should be reported. (optional)
                      Defaults to 'v2.0'.
    :param bool html: Add HTML describedby links to the structure.
    :param bool pdf: Add PDF describedby links to the structure.

    """

    _DESC_URL = 'https://developer.openstack.org/api-ref/identity/v2/'

    def __init__(self, href, id=None, html=True, pdf=True, **kwargs):
        super().__init__(id or 'v2.0', **kwargs)

        self.add_link(href)

        if html:
            self.add_html_description()
        if pdf:
            self.add_pdf_description()

    def add_html_description(self):
        """Add the HTML described by links.

        The standard structure includes a link to a HTML document with the
        API specification. Add it to this entry.
        """
        self.add_link(
            href=self._DESC_URL + 'content',
            rel='describedby',
            type='text/html',
        )

    def add_pdf_description(self):
        """Add the PDF described by links.

        The standard structure includes a link to a PDF document with the
        API specification. Add it to this entry.
        """
        self.add_link(
            href=self._DESC_URL + 'identity-dev-guide-2.0.pdf',
            rel='describedby',
            type='application/pdf',
        )


class V3Discovery(DiscoveryBase):
    """A Version element for a V3 identity service endpoint.

    Provides some default values and helper methods for creating a v3
    endpoint version structure. Clients should use this instead of creating
    their own structures.

    :param href: The url that this entry should point to.
    :param string id: The version id that should be reported. (optional)
                      Defaults to 'v3.0'.
    :param bool json: Add JSON media-type elements to the structure.
    :param bool xml: Add XML media-type elements to the structure.
    """

    def __init__(self, href, id=None, json=True, xml=True, **kwargs):
        super().__init__(id or 'v3.0', **kwargs)

        self.add_link(href)

        if json:
            self.add_json_media_type()
        if xml:
            self.add_xml_media_type()

    def add_json_media_type(self):
        """Add the JSON media-type links.

        The standard structure includes a list of media-types that the endpoint
        supports. Add JSON to the list.
        """
        self.add_media_type(
            base='application/json',
            type='application/vnd.openstack.identity-v3+json',
        )

    def add_xml_media_type(self):
        """Add the XML media-type links.

        The standard structure includes a list of media-types that the endpoint
        supports. Add XML to the list.
        """
        self.add_media_type(
            base='application/xml',
            type='application/vnd.openstack.identity-v3+xml',
        )


class DiscoveryList(dict[str, ty.Any]):
    """A List of version elements.

    Creates a correctly structured list of identity service endpoints for
    use in testing with discovery.

    :param string href: The url that this should be based at.
    :param bool v2: Add a v2 element.
    :param bool v3: Add a v3 element.
    :param string v2_status: The status to use for the v2 element.
    :param DateTime v2_updated: The update time to use for the v2 element.
    :param bool v2_html: True to add a html link to the v2 element.
    :param bool v2_pdf: True to add a pdf link to the v2 element.
    :param string v3_status: The status to use for the v3 element.
    :param DateTime v3_updated: The update time to use for the v3 element.
    :param bool v3_json: True to add a html link to the v2 element.
    :param bool v3_xml: True to add a pdf link to the v2 element.
    """

    TEST_URL = 'http://keystone.host:5000/'

    def __init__(
        self,
        href=None,
        v2=True,
        v3=True,
        v2_id=None,
        v3_id=None,
        v2_status=None,
        v2_updated=None,
        v2_html=True,
        v2_pdf=True,
        v3_status=None,
        v3_updated=None,
        v3_json=True,
        v3_xml=True,
    ):
        super().__init__(versions={'values': []})

        href = href or self.TEST_URL

        if v2:
            v2_href = href.rstrip('/') + '/v2.0'
            self.add_v2(
                v2_href,
                id=v2_id,
                status=v2_status,
                updated=v2_updated,
                html=v2_html,
                pdf=v2_pdf,
            )

        if v3:
            v3_href = href.rstrip('/') + '/v3'
            self.add_v3(
                v3_href,
                id=v3_id,
                status=v3_status,
                updated=v3_updated,
                json=v3_json,
                xml=v3_xml,
            )

    @property
    def versions(self):
        return self['versions']['values']

    def add_version(self, version):
        """Add a new version structure to the list.

        :param dict version: A new version structure to add to the list.
        """
        self.versions.append(version)

    def add_v2(self, href, **kwargs):
        """Add a v2 version to the list.

        The parameters are the same as V2Discovery.
        """
        obj = V2Discovery(href, **kwargs)
        self.add_version(obj)
        return obj

    def add_v3(self, href, **kwargs):
        """Add a v3 version to the list.

        The parameters are the same as V3Discovery.
        """
        obj = V3Discovery(href, **kwargs)
        self.add_version(obj)
        return obj

    def add_microversion(self, href, id, **kwargs):
        """Add a microversion version to the list.

        The parameters are the same as MicroversionDiscovery.
        """
        obj = MicroversionDiscovery(href=href, id=id, **kwargs)
        self.add_version(obj)
        return obj

    def add_nova_microversion(self, href, id, **kwargs):
        """Add a nova microversion version to the list.

        The parameters are the same as NovaMicroversionDiscovery.
        """
        obj = NovaMicroversionDiscovery(href=href, id=id, **kwargs)
        self.add_version(obj)
        return obj



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from keystoneauth1 import exceptions
from keystoneauth1 import loading
from keystoneauth1.tests.unit import utils as test_utils


class FedKerbLoadingTests(test_utils.TestCase):
    def test_options(self):
        opts = [
            o.name
            for o in loading.get_plugin_loader('v3fedkerb').get_options()
        ]

        allowed_opts = [
            'system-scope',
            'domain-id',
            'domain-name',
            'identity-provider',
            'project-id',
            'project-name',
            'project-domain-id',
            'project-domain-name',
            'protocol',
            'trust-id',
            'auth-url',
            'mutual-auth',
        ]

        self.assertCountEqual(allowed_opts, opts)

    def create(self, **kwargs):
        loader = loading.get_plugin_loader('v3fedkerb')
        return loader.load_from_options(**kwargs)

    def test_load_none(self):
        self.assertRaises(exceptions.MissingRequiredOptions, self.create)

    def test_load(self):
        self.create(
            auth_url='auth_url', identity_provider='idp', protocol='protocol'
        )



# Copyright 2018 SUSE Linux GmbH
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import typing as ty

from keystoneauth1.identity.v3 import base
from keystoneauth1 import session as ks_session


__all__ = ('ApplicationCredentialMethod', 'ApplicationCredential')


class ApplicationCredentialMethod(base.AuthMethod):
    """Construct a User/Passcode based authentication method.

    :param string application_credential_secret: Application credential secret.
    :param string application_credential_id: Application credential id.
    :param string application_credential_name: The name of the application
                                               credential, if an ID is not
                                               provided.
    :param string username: Username for authentication, if an application
                            credential ID is not provided.
    :param string user_id: User ID for authentication, if an application
                           credential ID is not provided.
    :param string user_domain_id: User's domain ID for authentication, if an
                                  application credential ID is not provided.
    :param string user_domain_name: User's domain name for authentication, if
                                    an application credential ID is not
                                    provided.
    """

    application_credential_secret: str
    application_credential_id: str | None = None
    application_credential_name: str | None = None
    user_id: str | None = None
    username: str | None = None
    user_domain_id: str | None = None
    user_domain_name: str | None = None

    def __init__(
        self,
        *,
        application_credential_secret: str,
        application_credential_id: str | None = None,
        application_credential_name: str | None = None,
        user_id: str | None = None,
        username: str | None = None,
        user_domain_id: str | None = None,
        user_domain_name: str | None = None,
    ) -> None:
        self.application_credential_secret = application_credential_secret
        self.application_credential_id = application_credential_id
        self.application_credential_name = application_credential_name
        self.user_id = user_id
        self.username = username
        self.user_domain_id = user_domain_id
        self.user_domain_name = user_domain_name

    def get_auth_data(
        self,
        session: ks_session.Session,
        auth: base.Auth,
        headers: dict[str, str],
        request_kwargs: dict[str, object],
    ) -> tuple[None, None] | tuple[str, ty.Mapping[str, object]]:
        auth_data: dict[str, ty.Any] = {
            'secret': self.application_credential_secret
        }

        if self.application_credential_id:
            auth_data['id'] = self.application_credential_id
        else:
            auth_data['name'] = self.application_credential_name
            auth_data['user'] = {}
            if self.user_id:
                auth_data['user']['id'] = self.user_id
            elif self.username:
                auth_data['user']['name'] = self.username

                if self.user_domain_id:
                    auth_data['user']['domain'] = {'id': self.user_domain_id}
                elif self.user_domain_name:
                    auth_data['user']['domain'] = {
                        'name': self.user_domain_name
                    }

        return 'application_credential', auth_data

    def get_cache_id_elements(self) -> dict[str, str | None]:
        return {
            'application_credential_application_credential_secret': self.application_credential_secret,
            'application_credential_application_credential_id': self.application_credential_id,
            'application_credential_application_credential_name': self.application_credential_name,
            'application_credential_user_id': self.user_id,
            'application_credential_username': self.username,
            'application_credential_user_domain_id': self.user_domain_id,
            'application_credential_user_domain_name': self.user_domain_name,
        }


class ApplicationCredential(base.Auth):
    """A plugin for authenticating with an application credential.

    :param string auth_url: Identity service endpoint for authentication.
    :param string application_credential_secret: Application credential secret.
    :param string application_credential_id: Application credential ID.
    :param string application_credential_name: Application credential name.
    :param string username: Username for authentication.
    :param string user_id: User ID for authentication.
    :param string user_domain_id: User's domain ID for authentication.
    :param string user_domain_name: User's domain name for authentication.
    :param bool reauthenticate: Allow fetching a new token if the current one
                                is going to expire. (optional) default True
    """

    _auth_method_class = ApplicationCredentialMethod

    def __init__(
        self,
        auth_url: str,
        application_credential_secret: str,
        application_credential_id: str | None = None,
        application_credential_name: str | None = None,
        user_id: str | None = None,
        username: str | None = None,
        user_domain_id: str | None = None,
        user_domain_name: str | None = None,
        *,
        unscoped: bool = False,
        trust_id: str | None = None,
        system_scope: str | None = None,
        domain_id: str | None = None,
        domain_name: str | None = None,
        project_id: str | None = None,
        project_name: str | None = None,
        project_domain_id: str | None = None,
        project_domain_name: str | None = None,
        reauthenticate: bool = True,
        include_catalog: bool = True,
    ) -> None:
        method = self._auth_method_class(
            application_credential_secret=application_credential_secret,
            application_credential_id=application_credential_id,
            application_credential_name=application_credential_name,
            user_id=user_id,
            username=username,
            user_domain_id=user_domain_id,
            user_domain_name=user_domain_name,
        )
        super().__init__(
            auth_url,
            [method],
            unscoped=unscoped,
            trust_id=trust_id,
            system_scope=system_scope,
            domain_id=domain_id,
            domain_name=domain_name,
            project_id=project_id,
            project_name=project_name,
            project_domain_id=project_domain_id,
            project_domain_name=project_domain_name,
            reauthenticate=reauthenticate,
            include_catalog=include_catalog,
        )
